package decatur.de.rigorous.ivp;
import decatur.de.rigorous.numutil.*;


/* Define structure which holds all relevant information about a d*m-zonotope.
   Variables of type zonotope always point to such a structure.
   m is the order of the d*m-zonotope.
   The m d-zonotopes are represented by R[0]...R[m-2] (m-1 "full" d-zonotopes)
   and one interval I.
   0<=k<=m-2 is the first non-zero d-zonotope, i.e.
   Z = I+R[k]+...+R[m-2].

/----\/----\/----\/----\/----\/----\/----\/----\
| I  || R0 || R1 || .. || Rk || .. ||    ||Rm-2|
\----/\----/\----/\----/\----/\----/\----/\----/

   If k==m-1, then the zonotope is simply I.
   A, T and r are auxilary quantities.  */



   
public class Zonotope {
  int m, k, dim, debug;
  boolean do_hull;
  double hull[];
  double radii[];
  double I[];
  double [][][] R;
  double [][]T;
  Interval [][]A;
  double []r;
  double[][] line;
  private double avg_alpha=1.0, avg_delta=0.0;
  boolean optimized = true;

/* Allocate and return an dim*m-zonotope.  */
  Zonotope (int dim, int m) {
    int i;
    this.m = m;
    this.dim = dim;
    debug = 0;
    do_hull = true;
    hull = new double[dim];
    I = new double[dim];
    A  = Interval.matrix(dim);
    T  = new double[dim][dim];
    r  = new double[dim];
    radii  = new double[m];
    R  = new double[m-1][dim][dim];
  }

 
/* Copy a zonotope of order m to a zonotope of order m+1.

/----\/----\/----\/----\/----\/----\/----\/----\
| I  || R0 || R1 || .. || Rk || .. ||    ||Rm-2|
\----/\----/\----/\----/\----/\----/\----/\----/
   |                        \__
   |                           \  ......
   |                            \
/----\/----\/----\/----\/----\/----\/----\/----\/----\
| I  ||    ||    ||    ||    ||    ||    ||    ||    |
\----/\----/\----/\----/\----/\----/\----/\----/\----/

*/
  void copy (Zonotope z) {
    k=z.k+1;
    blas.copy (I, z.I);
    for (int i=k; i<=m-2; i++)
      blas.copy (R[i], z.R[i-1]); 
    do_hull = z.do_hull;
    avg_delta = z.avg_delta;
    optimized = z.optimized;
    if ( !do_hull ) 
	blas.copy (hull, z.hull);
    getRadii(radii);
  }

/* Initialize a zonotope z, (which is already allocated). The zonotope
   will be set to an interval of radius rad.  */
  public void init (double []rad) {
  /* Only one d-zonotope is used, which is I.  */
    k=m-1;
    blas.copy(I, rad);
    do_hull = true;
    avg_delta=0.0;
    getRadii(radii);
  }


/* Initialize a zonotope z (which is already allocated). The zonotope
   will be set to a d-zonotope represented by matrix A.  */
  public void init (double [][]A) {
  /* Initialize d-zonotopes I and R(k=2).  */
    k=m-2;
    blas.flush(I);
    blas.copy(R[m-2], A);
    do_hull = true;
    avg_delta=0.0;
    getRadii(radii);
  }

  public void debug (int debug_level) {
    debug = debug_level;
  }

/* Querry the interval hull of a zonotope.  */
  public double[] ihull() {
    if ( !do_hull ) return hull;
    blas.copy (hull, I);
    for (int i=k; i<=m-2; i++) {
      blas.row_sum (r, R[i]); 
      blas.add (hull, hull, r);
    }
    do_hull = false;
    return hull;
  }
  
  public double[][] getLines () {
    return line;
  }

  public double preparePlot () {
    double rad = blas.lsup_norm (ihull());
    line = new double[(m-k)*dim][dim];
    int l = 0;
    for (int j=0; j<dim; j++) {
      line[l] = new double[dim];
      line[l++][j] = I[j];
    }
    
    for (int i=k; i<=m-2; i++)
      for (int j=0; j<dim; j++) {
	line[l] = new double[dim];
	for (int n=0; n<dim; n++)
	  line[l][n] = R[i][n][j];
	l++;
      }
    int n = line.length;
    double angle[] = new double[n];
    
    for (int i=0; i<n; i++) {
      if ( line[i][1]<0 ) {
	line[i][0] *= -1;
	line[i][1] *= -1;
      }
      angle[i] = Math.atan2(line[i][0], line[i][1]);
    }
    double temp[] = new double[2];  
    for (int i=1; i<n; i++) {
      int j;
      double a = angle[i];
      blas.copy (temp, line[i]);
      for (j=i; j>0; j--) {
	if ( a>angle[j-1] ) break;
	angle[j] = angle[j-1];
	blas.copy (line[j], line[j-1]);
      }
      angle [j] = a;
      blas.copy (line[j], temp);
    }
    return rad;
  }
  
  public double getAverage() {
    return avg_delta;
  }

  void getRadii(double []dest) {
    double d;
    for (int i=0; i<k; i++) dest[i] = 0;
    dest[k] = blas.l1_norm (I);
    for (int i=k; i<=m-2; i++) {
      blas.row_sum (r, R[i]); 
      dest[i+1] = blas.l1_norm (r);
    }
  }

    public void setOptimized (boolean on) {
	optimized = on;
    }

    private boolean query1 (double a, double b) {
	return  a*Math.log(a/avg_delta)/Math.log(m) > b;
    }

    private boolean query0 (double a, double b) {
	return  a > b;
    }

  /* Primary Cascade algorithm. Computes zonotope enclosure of rad+Sz.
     Here rad is the inflating interval.  */
  public void eval (double [][] S, double [] rad) {
    do_hull = true;
    if ( debug==1 ) System.out.println("zono: "+k);
    if ( debug==2 ) System.out.println("zono: ");
    
    avg_delta = (blas.l1_norm(rad)+9*avg_delta)/10.0;
    
    /* Initialize with the inflating interval.  */
    
    for (int j=0; j<dim; j++)
      for (int l=0; l<dim; l++)
	A[j][l] = Interval.mul(S[j][l], I[l]);

    blas.abs(T, A);
    
    double d = blas.l1_norm (I);
    blas.row_sum(I, T);
    
    if ( k>0 ) {
      blas.row_sum(r, T);
      radii[k]=blas.l1_norm (r);

      blas.copy (I, rad);
      for (int i=k-1; i<=m-2; i++ ) {
	if ( i!=k-1 ) blas.mul (A, S, R[i]);
	//	System.out.println("\nk:"+k);
	//blas.print(R[i]); blas.print(S);
	blas.Interval2center (R[i], T, A);
	//	blas.print(R[i]);
	blas.row_sum(r, T);
	blas.add (I, I, r);
      }
      k--;
    } else {
      blas.row_sum(I, T);
      blas.add (I, I, rad);

      double sum = blas.l1_norm (I);
      int i, maxi = 0;
      for (int j=1; j<=m-1; j++) {
	  boolean b;
	  if ( optimized )
	      b = query1 (sum, radii[j]);
	  else
	      b = query0 (sum, radii[j]);

	  if ( b )
	      maxi = j;
	  sum += radii[j];
      }
      
      for (i=1; i<=maxi; i++ ) {
	blas.mul (A, S, R[i-1]);
	blas.abs(T, A);
	blas.row_sum(r, T);
	blas.add (I, I, r);
	radii[i] = 0.0;
      }
      radii[maxi] = blas.l1_norm (I);
      for (i=maxi+1; i<=m-1; i++ ) {
	blas.mul (A, S, R[i-1]);
	blas.Interval2center (R[i-1], T, A);
	blas.row_sum(r, T);
	blas.add (I, I, r);
      }
      k = maxi;
    }
  }

  /* Interface to the primary Cascade algorithm in case the mapping is
     not "thin". Ac is its center, Ar is its radius. Note that
     zono(S, ZERO_MATRIX, a, z)=zono_prim(S, a, z).  */
  public void eval (double [][]Ac, double [][]Ar, double []a) {
    if ( do_hull ) hull = ihull();
    blas.mul (r, Ar, hull);
    blas.add(a, a, r);
    eval (Ac, a);
  }
}


