package decatur.de.rigorous.numutil;

public class Interval {
    public double inf, sup;

    public Interval (Interval src) {
	set (src);
    }
    
    public Interval (double inf, double sup) {
	set (inf, sup);
    }
    
    public Interval (double c) {
	set (c);
    }
    
    public Interval () {
	set (0.0);
    }
    
    public String toString () {
	return "["+String.valueOf(this.inf)+", "+String.valueOf(this.sup)+"]";
    }
    
    public void parseString (String s) throws NumberFormatException {
	String msg = new String (" expected reading Interval");
	int open = s.indexOf ('[');
	
	//	System.out.println ("\n"+s);

	if ( open == -1 )
	    throw new NumberFormatException("\n'['"+msg);

	int comma = s.indexOf (',');
	if ( comma == -1 )
	    throw new NumberFormatException("\n','"+msg);

	int close = s.indexOf (']');
	if ( close == -1 )
	    throw new NumberFormatException("\n']'"+msg);	

	inf = Double.valueOf(s.substring (open +1, comma)).doubleValue();
	sup = Double.valueOf(s.substring (comma+1, close)).doubleValue();
    }

    public void set (Interval src) {
	inf = src.inf;
	sup = src.sup;
    }
    
    public void set (double src) {
	inf = sup = src;
    }
    
    public void set (double inf, double sup) {
	if ( inf>sup )
	    throw new ArithmeticException ("Interval bounds inconsistent: inf="+
					   inf+" sup="+sup);
	this.inf = inf; this.sup = sup;
    }
    
    public void setCentered (double c, double r) {
	inf=IEEE.pred(c-r);
	sup=IEEE.succ(c+r);
    }
    
    public CenteredInterval centered () {
	CenteredInterval x = new CenteredInterval ();
	x.center = (sup+inf)/2;
	x.radius = Math.max(IEEE.succ(sup-x.center),
			    IEEE.succ(x.center-inf));
	return x;
    }
    
    public double mid () {
	return (sup+inf)/2.0;
    }
    
    public double rad () {
	return IEEE.succ((sup-inf)/2.0);
    }
    
    public double width () {
	return IEEE.succ(sup-inf);
    }
    
    public void inflate (double eps) {
	eps = Math.abs (eps);
	inf = IEEE.pred (inf-eps);
	sup = IEEE.succ (sup+eps);
    }

    public boolean contains (Interval a) {
	return (inf<=a.inf && sup>=a.sup);
    }
    
    public boolean contains (double a) {
	return (inf<=a && sup>=a);
    }
    
  public void add (Interval a) {
      inf=IEEE.pred(inf+a.inf);
      sup=IEEE.succ  (sup+a.sup);
  }
    
    public void sub (Interval a) {
	inf=IEEE.pred(inf-a.sup);
	sup=IEEE.succ  (sup-a.inf);
    }
    
    public void mul (Interval y) {
	set(Interval.mul(this, y));
    }
    
    public Interval mul (double c) {
	if (c>=0.0) { 
	    inf = IEEE.pred (inf*c);
	    sup = IEEE.succ (sup*c); 
	} else {
	    double d = IEEE.pred (sup*c);
	    sup = IEEE.succ (inf*c);
	    inf = d;
	}
	return this;
    }

    public void inv () {
	if ( inf<=0.0 && sup>=0.0 )
      throw new ArithmeticException ("Divison by zero");
	set (IEEE.pred(1/sup), IEEE.succ(1/inf));
    }

    public void div (Interval y) {
	Interval w = new Interval(y);
	w.inv();  mul(w);
    }
    
    public void neg () {
    double t = inf;
    inf = -sup;
    sup = -t;
    }

    double powerUp (double b, int exp) {
	double r = b;
	for (int k=1; k<exp; k++)
	    r = IEEE.succ (r*b);
	return r;
    }

    double powerDown (double b, int exp) {
	double r = b;
	for (int k=1; k<exp; k++)
	    r = IEEE.pred (r*b);
	return r;
    }
    
    public void pow (int exp) {
	if ( exp==0 )
	    this.set (Interval.ZERO);
	else if ( exp < 0 ) {
	    inv ();
	    pow (-exp);
	} else {
	    if ( inf>0.0 )
		set(powerDown(inf,exp), powerUp(sup,exp));
	    else if ( sup<0.0 ) {
		if ( exp%2==0 )
		    set(powerDown(-sup,exp), powerUp(-inf,exp));
		else
		    set(-powerUp(-inf,exp), -powerDown(-sup,exp));
	    } else {		// inf<0<sup
		if ( exp%2==0 )
		    set(0.0, powerUp(Math.max(Math.abs(inf), Math.abs(sup)), exp));
		else
		    set (-powerUp(-inf,exp), powerUp(sup, exp)); 
	    }
	}
    }
    
  public void sqrt () {
    if ( inf<0.0 ) 
      throw new ArithmeticException ("Negative arg to square root");
    inf = IEEE.pred(Math.sqrt(inf));
    sup = IEEE.succ(Math.sqrt(sup));
  }

  static final double Pi         = 3.14159265358979323846;
  static final double TwoPi      = 6.28318530717958647693;
  static final double PiHalf     = 1.57079632679489661923;
  static final double PiQuarter  = 0.78539816339744830962;
  

  public void sin () {
    double inf_y=-1, sup_y=1, q;
    int s1, s2;
    
    if ( rad() >= Pi) {
      set (-1.0, 1.0);
      return;
    }
    q = Math.floor(IEEE.pred(inf/TwoPi));
    q *= TwoPi;
    if ( inf<=q+PiHalf || inf>=q+3*PiHalf ) s1=1;
    else s1=-1;

    q = Math.floor(IEEE.pred(sup/TwoPi));
    q *= TwoPi;
    if ( sup<=q+PiHalf || sup>=q+3*PiHalf ) s2=1;
    else s2=-1;

    if ( s1>=0 && s2>=0 ) {
      if ( sup-inf<=PiHalf ) {
	inf_y = Math.sin(inf);
	sup_y = Math.sin(sup);
	inf_y = IEEE.pred(inf_y);
	sup_y = IEEE.succ(sup_y);
      } else {
	inf_y=-1; sup_y=1;
      }
    } else if (s1<=0 && s2<=0) {
      if (sup-inf<=PiHalf) {
	inf_y = Math.sin(sup);
	sup_y = Math.sin(inf);
	inf_y = IEEE.pred(inf_y);
	sup_y = IEEE.succ(sup_y);
      } else {
	inf=-1; sup=1;
      }
    } else if (s1>=0 && s2<=0) {
      inf_y = Math.min(Math.sin(inf),Math.sin(sup));
      inf_y = IEEE.pred(inf_y);
      sup_y = 1.0;
    } else if (s1<=0 && s2>=0) {
      sup_y = Math.max(Math.sin(inf),Math.sin(sup));
      sup_y = IEEE.succ(sup_y);
      inf_y = -1.0;
    }
    if (inf_y < -1.0) inf_y = -1.0;
    if (sup_y > 1.0) sup_y = 1.0;
    set(inf_y, sup_y);
  }

  public void cos () {
    Interval PiHalfIncl = new Interval(IEEE.pred(PiHalf), IEEE.succ(PiHalf));
    add (PiHalfIncl);
    sin ();
  }

  public void exp () {
    inf = Math.exp (inf);
    sup = Math.exp (sup);
    inf = IEEE.down (inf);
    sup = IEEE.up   (sup);
    if (inf < 0.0) inf = 0.0;
}

  public void ln () {
    if ( inf<0.0 ) 
      throw new ArithmeticException ("Negative arg to ln()");
    inf = IEEE.pred(Math.log(inf));
    sup = IEEE.succ(Math.log(sup));
  }

  public static Interval ZERO = new Interval(0);
  public static Interval ONE  = new Interval(1);
  public static Interval TWO  = new Interval(2);


  public static Interval []vector (int dim) {
    Interval []x = new Interval[dim];
    for (int i=0; i<dim; i++) x[i]=new Interval(0);
    return x;
  }

  public static Interval [][]matrix (int dim) {
    Interval [][]X = new Interval[dim][dim];
    for (int i=0; i<dim; i++)
      for (int j=0; j<dim; j++)
	X[i][j]=new Interval(0);
    return X;
  }

  public static Interval mul(double a, double b) {
    a *= b;
    return new Interval(IEEE.pred(a), IEEE.succ(a));
  }

  public static Interval neg(Interval a) {
    return new Interval(-a.sup, -a.inf);
  }

  public static Interval add(Interval a, Interval b) {
    Interval c = new Interval(a);
    c.add (b);
    return c;
  }

  public static Interval sub(Interval a, Interval b) {
    return new Interval(IEEE.pred(a.inf-b.sup), IEEE.succ(a.sup-b.inf));
  }

  public static Interval mul(Interval x, Interval y) {
  Interval w = new Interval(0);
  if  (x.inf>=0.0) {
    if (y.inf>=0.0) { 
      w.inf = IEEE.pred(x.inf*y.inf); /* 1 */
      w.sup = IEEE.succ  (x.sup*y.sup); 
    }
    else if ((y.inf<0.0) && (y.sup>0.0)) {
      w.inf = IEEE.pred(x.sup*y.inf); /* 4 */
      w.sup = IEEE.succ  (x.sup*y.sup); }
    else {
      w.inf = IEEE.pred(x.sup*y.inf); /* 6 */
      w.sup = IEEE.succ  (x.inf*y.sup); }
  }
  else if (x.sup<=0.0) {
    if (y.inf>=0.0) {
      w.inf = IEEE.pred(x.inf*y.sup); /* 3 */
      w.sup = IEEE.succ  (x.sup*y.inf); }
    else if ((y.inf<0.0) && (y.sup>0.0)) {
      w.inf = IEEE.pred(x.inf*y.sup); /* 5 */
      w.sup = IEEE.succ  (x.inf*y.inf); 
    }
    else {
      w.inf = IEEE.pred(x.sup*y.sup); /* 8 */
      w.sup = IEEE.succ  (x.inf*y.inf); }
  }
  else {
    if (y.inf>=0.0) {
      w.inf = IEEE.pred(x.inf*y.sup); /* 2 */
      w.sup = IEEE.succ  (x.sup*y.sup); }
    else if (y.sup<=0.0) {
      w.inf = IEEE.pred(x.sup*y.inf); /* 7 */
      w.sup = IEEE.succ  (x.inf*y.inf); 
    }
    else {
      double a,b;
      a = IEEE.pred(x.inf*y.sup);
      b = IEEE.pred(x.sup*y.inf);
      w.inf = Math.min(a,b); /* 9 */
      a = IEEE.succ(x.inf*y.inf);
      b = IEEE.succ(x.sup*y.sup);
      w.sup = Math.max(a,b); 
    }
  }
  return w;
  }

  public static double abs(Interval a) {
    return Math.max(Math.abs(a.inf), Math.abs(a.sup));
  }

}

